
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <libssp.h>

#include <wiringPi.h>

#include <signal.h>

#include<unistd.h>

#include<opencv/cv.h>
#include <opencv/highgui.h>
// sudo apt-get install libopencv-dev
// sudo apt-get install libgl1-mesa-dri

#include <sys/time.h>

#include<pthread.h>

static int frame_count[2];

volatile int stop_process;
static void signal_handler(int sig)
{
  printf("Terminated.\n");
  if(sig == SIGINT)
    stop_process = 1;
}


void frame_drop_func(struct ssp_handle *handle){

    printf("Frame dropped by camera.\n");

}

void frame_drop_preprocess_func(struct ssp_handle *handle){

  printf("Frame dropped at preprocess.\n");
}

IplImage* BayerImg[2], *RGBImg[2];

volatile int ActiveCamNo = 0;

void GammaCorrection(IplImage* src, IplImage* dst, double gamma){
  int i;
  uchar LUT[256];

  for (i = 0; i < 256; i++){
    LUT[i] = (int)(pow((double)i / 255.0, 1.0 / gamma) * 255.0);
  }
  CvMat lut_mat = cvMat(1, 256, CV_8UC1, LUT);
  cvLUT(src, dst, &lut_mat);
}

volatile int SettingMode = 0;
#define GAIN_MODE 0
#define SS_MODE 1 // Shutter Speed
volatile int GammaEnable = 0;

struct __paracopy_args {
  int width;
  int height;
  unsigned char *src;
  unsigned char *dest;
};

void *paracopy_thread_func(void *param){
  struct __paracopy_args *copyparams = (struct __paracopy_args *)param;
  int i;
  for(i=0;i<copyparams->height;i++)
    memcpy(copyparams->dest+copyparams->width*i,
	   copyparams->src+copyparams->width*i,
	   copyparams->width);

  return NULL;
}

char *model_name[2];

pthread_mutex_t frame_mutex;

#define GAIN_CHANGE_VAL 25

#define NUM_COPY_THREAD 4
pthread_t thread_handle[2][NUM_COPY_THREAD];
struct __paracopy_args copy_params[2][NUM_COPY_THREAD];

void fame_preprocess(struct ssp_handle *handle, struct ssp_frame *frame){
  //  printf("Preprocess is active. frame_count = %d\n", frame_count);
  //if(frame_count % 10 == 0) printf("frame_count = %d\n", frame_count);

  pthread_mutex_lock(&frame_mutex);
  int i,j;
  int CamNo = ssp_get_frame_CamNo(frame);
  int Width = ssp_get_frame_width(frame);
  int Height = ssp_get_frame_height(frame);
  unsigned char *frame_data = ssp_get_frame_data(frame);

#if 0 // Simple copy
  for(i=0;i<BayerImg[CamNo]->height;i++)
    for(j=0;j<BayerImg[CamNo]->width;j++){
      ((unsigned char *)(BayerImg[CamNo]->imageData + BayerImg[CamNo]->widthStep * i))[j] = frame_data[i*BayerImg[CamNo]->widthStep+j];
    }
#endif

#if 0 // memcpy
  for(i=0;i<BayerImg[CamNo]->height;i++){
    memcpy((unsigned char *)(BayerImg[CamNo]->imageData + BayerImg[CamNo]->widthStep * i), &frame_data[i*BayerImg[CamNo]->widthStep], BayerImg[CamNo]->width);
  }
#endif
#if 0 // long long copy
  for(i=0;i<BayerImg[CamNo]->height;i++)
    for(j=0;j<BayerImg[CamNo]->width/sizeof(long long);j++){
      *((long long *)(BayerImg[CamNo]->imageData + BayerImg[CamNo]->widthStep * i) + j) = *((long long*)(frame_data+i*BayerImg[CamNo]->widthStep)+j);
    }
#endif
#if 1 // parallel copy
  for(i=0;i<NUM_COPY_THREAD;i++){
    copy_params[CamNo][i].width = BayerImg[CamNo]->widthStep;
    copy_params[CamNo][i].height = BayerImg[CamNo]->height/NUM_COPY_THREAD;
    copy_params[CamNo][i].dest = (unsigned char *)(BayerImg[CamNo]->imageData + BayerImg[CamNo]->widthStep * (i*BayerImg[CamNo]->height/NUM_COPY_THREAD));
    copy_params[CamNo][i].src = &frame_data[(i*BayerImg[CamNo]->height/NUM_COPY_THREAD)*BayerImg[CamNo]->widthStep];
    pthread_create(&(thread_handle[CamNo][i]) , NULL , paracopy_thread_func ,&(copy_params[CamNo][i]));
      
  }
  for(i=0;i<NUM_COPY_THREAD;i++)
    pthread_join(thread_handle[CamNo][i] , NULL);
#endif

    ssp_release_frame(frame);

    cvCvtColor( BayerImg[CamNo], RGBImg[CamNo], CV_BayerRG2RGB);
    if(GammaEnable == 1)
      GammaCorrection(RGBImg[CamNo], RGBImg[CamNo], 2.2);

    char WindowName[256];

    if(ActiveCamNo == CamNo){
      sprintf(WindowName, "SSP Viewer");
      cvShowImage(WindowName, RGBImg[CamNo]);
    }


    int KeyCode = cvWaitKey(2);
    KeyCode &= 0xFF;
    if(KeyCode == 'q'){
      stop_process=1;
      printf("Stop process requested\n");
    }
    if(KeyCode == 'c'){
      ActiveCamNo= ActiveCamNo == 0 ? 1 : 0;
      printf("Active cam changed to %d\n", ActiveCamNo);
    }
    if(KeyCode == 's'){
      printf("Setting Mode is set to Shutter Speed Mode.\n");
      SettingMode = SS_MODE;
    }
    if(KeyCode == 'g'){
      printf("Setting Mode is set to Gain Mode.\n");
      SettingMode = GAIN_MODE;
    }
    if(KeyCode == '0'){
      if(GammaEnable == 0){
	GammaEnable = 1;
	printf("Gamma Correction is Enabled.\n");
      }
      else{
	GammaEnable = 0;
	printf("Gamma Correction is Disabled.\n");
      }
    }
    if(KeyCode == '+' || KeyCode == '-'){
      int RegVal;

      
      ssp_control_focus_on_camera(handle);
      
      if(SettingMode == GAIN_MODE){
	if(strcmp(model_name[CamNo], "IMX219") == 0){
	  if(__CCIRegRead(handle, 0x0157, &RegVal) == 0)
	    printf("Reading register error.\n");
	  if(KeyCode == '+'){
	    RegVal += GAIN_CHANGE_VAL;
	    if(RegVal >= 233) RegVal = 0;
	  }
	  if(KeyCode == '-'){
	    RegVal -= GAIN_CHANGE_VAL;
	    if(RegVal <= -1) RegVal = 232;
	  }
	  if(__CCIRegWrite(handle, 0x0157, RegVal) == 0)
	    printf("Writing register error.\n");
	  if(__CCIRegRead(handle, 0x0157, &RegVal) == 0)
	    printf("Reading register error.\n");
	  printf("Analog Gain Set to %d\n", RegVal);

	}
	if(strcmp(model_name[CamNo], "IU233") == 0){
	  if(__CCIRegRead(handle, 0x0205, &RegVal) == 0)
	    printf("Reading register error.\n");
	  if(KeyCode == '+'){
	    RegVal += GAIN_CHANGE_VAL;
	    if(RegVal >= 225) RegVal = 0;
	  }
	  if(KeyCode == '-'){
	    RegVal -= GAIN_CHANGE_VAL;
	    if(RegVal <= -1) RegVal = 224;
	  }
	  if(__CCIRegWrite(handle, 0x0205, RegVal) == 0)
	    printf("Writing register error.\n");
	  if(__CCIRegRead(handle, 0x0205, &RegVal) == 0)
	    printf("Reading register error.\n");
	  printf("Analog Gain Set to %d\n", RegVal);

	}
	if(strcmp(model_name[CamNo], "IMX408") == 0){
	  int GainVal;
	  if(__CCIRegRead(handle, 0x0205, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal = RegVal;
	  if(__CCIRegRead(handle, 0x0204, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal |= ((RegVal & 0x1) << 8);
	  RegVal = GainVal;
	  if(KeyCode == '+'){
	    RegVal += GAIN_CHANGE_VAL;
	    if(RegVal >= 481) RegVal = 0;
	  }
	  if(KeyCode == '-'){
	    RegVal -= GAIN_CHANGE_VAL;
	    if(RegVal <= -1) RegVal = 480;
	  }
	  if(__CCIRegWrite(handle, 0x0205, RegVal & 0xFF) == 0)
	    printf("Writing register error.\n");
	  if(__CCIRegWrite(handle, 0x0204, (RegVal>>8)&0x1) == 0)
	    printf("Writing register error.\n");	
	  if(__CCIRegRead(handle, 0x0205, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal = RegVal;
	  if(__CCIRegRead(handle, 0x0204, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal |= ((RegVal & 0x1) << 8);
	  RegVal = GainVal;
	  printf("Analog Gain Set to %d\n", RegVal);

	}
	if(strcmp(model_name[CamNo], "IMX378") == 0){
	  int GainVal;
	  if(__CCIRegRead(handle, 0x0205, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal = RegVal;
	  if(__CCIRegRead(handle, 0x0204, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal |= ((RegVal & 0x3) << 8);
	  RegVal = GainVal;
	  if(KeyCode == '+'){
	    RegVal += GAIN_CHANGE_VAL;
	    if(RegVal >= 978) RegVal = 0;
	  }
	  if(KeyCode == '-'){
	    RegVal -= GAIN_CHANGE_VAL;
	    if(RegVal <= -1) RegVal = 978;
	  }
	  if(__CCIRegWrite(handle, 0x0205, RegVal & 0xFF) == 0)
	    printf("Writing register error.\n");
	  if(__CCIRegWrite(handle, 0x0204, (RegVal>>8)&0x3) == 0)
	    printf("Writing register error.\n");	
	  if(__CCIRegRead(handle, 0x0205, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal = RegVal;
	  if(__CCIRegRead(handle, 0x0204, &RegVal) == 0)
	    printf("Reading register error.\n");
	  GainVal |= ((RegVal & 0x3) << 8);
	  RegVal = GainVal;
	  printf("Analog Gain Set to %d\n", RegVal);

	}
      }

      if(SettingMode == SS_MODE){
	if(strcmp(model_name[CamNo], "IMX219") == 0){
	  int frame_length_lines;
	  int coarse_integration_time;
	  
	  if(__CCIRegRead(handle, 0x0160, &RegVal) == 0)
	    printf("Reading register error.\n");
	  frame_length_lines = RegVal<<8;
	  if(__CCIRegRead(handle, 0x0161, &RegVal) == 0)
	    printf("Reading register error.\n");
	  frame_length_lines |= RegVal;

	  if(__CCIRegRead(handle, 0x015A, &RegVal) == 0)
	    printf("Reading register error.\n");
	  coarse_integration_time = RegVal<<8;
	  if(__CCIRegRead(handle, 0x015B, &RegVal) == 0)
	    printf("Reading register error.\n");
	  coarse_integration_time |= RegVal;

	  if(KeyCode == '+'){
	    coarse_integration_time ++;
	    if(coarse_integration_time == (frame_length_lines-5)) RegVal = 1;
	  }
	  if(KeyCode == '-'){
	    coarse_integration_time --;
	    if(coarse_integration_time == -1) coarse_integration_time = frame_length_lines-4;
	  }
	  if(__CCIRegWrite(handle, 0x015A, (coarse_integration_time>>8)&0xFF) == 0)
	    printf("Writing register error.\n");
	  if(__CCIRegWrite(handle, 0x015B, coarse_integration_time&0xFF) == 0)
	    printf("Writing register error.\n");
	  if(__CCIRegRead(handle, 0x015A, &RegVal) == 0)
	    printf("Reading register error.\n");
	  coarse_integration_time = RegVal<<8;
	  if(__CCIRegRead(handle, 0x015B, &RegVal) == 0)
	    printf("Reading register error.\n");
	  coarse_integration_time |= RegVal;
	  printf("Shutter Speed Set (coarse_integration_time) to %d\n", coarse_integration_time);
	}
      }
    }
    
    printf("Cam %d : frame  count = %d\n", CamNo,frame_count[CamNo]);
    frame_count[CamNo] ++;

  pthread_mutex_unlock(&frame_mutex);
}

int __CCIRegRead(struct ssp_handle *handle, int address, int *data);
int __CCIRegWrite(struct ssp_handle *handle, int address, int data);

#define GAMMA_OPTION SSP_DISABLE
//#define GAMMA_OPTION SSP_ENABLE

struct __cam_thread_args {
  char *profile_filename;
  char *model_name;
  char CamNo;
};

pthread_mutex_t cam_mutex;

void cam_thread_func(void *param){

  struct __cam_thread_args *cam_param = (struct __cam_thread_args *)param;

  struct ssp_settings settings;

  settings.lib_settings.NumFrameFIFOSize = 2;
  settings.lib_settings.NumThreadForBuildInPreprocess = 4;
  settings.lib_settings.frame_drop_cam_user_func = frame_drop_func;
  settings.lib_settings.frame_drop_pre_user_func = frame_drop_preprocess_func;
  settings.lib_settings.fame_preprocess_user_func = fame_preprocess;
  settings.frame_preprocess_options.FormatConvertType = SSP_FRAME_BAYER8;
  settings.frame_preprocess_options.GammaCorrectionEnable = GAMMA_OPTION;
  settings.frame_preprocess_options.GammaVal = 2.2;
  settings.camera_settings.PowerOnResetUsec = 1000000;
  settings.camera_settings.CameraNo = cam_param->CamNo;
  
  struct ssp_handle *handle;

  struct ssp_profile *profile;
  char *profile_filename = cam_param->profile_filename;
  
  ssp_read_profile(profile_filename, &profile);

  printf("Setting summary: %d x %d @ %d fps %d bits\n",
	 profile->ImageProperty.Width,
	 profile->ImageProperty.Height,
	 profile->ImageProperty.Frequency,
	 profile->ImageProperty.BayerBits);


  // OpenCV Settings
  char WindowName[256];

  if(cam_param->CamNo == ActiveCamNo){
    sprintf(WindowName, "SSP Viewer");
    printf("%s\n", WindowName);
    cvNamedWindow(WindowName, CV_WINDOW_AUTOSIZE); 
    printf("OK\n");
  }

    
  BayerImg[cam_param->CamNo] = cvCreateImage(cvSize(profile->ImageProperty.Width,profile->ImageProperty.Height),IPL_DEPTH_8U,1);
  RGBImg[cam_param->CamNo] = cvCreateImage(cvSize(profile->ImageProperty.Width,profile->ImageProperty.Height),IPL_DEPTH_8U,3);

    
  int i,j;
  for(i=0;i<1;i++){ // for repeat initialization test.
    int retval;
    pthread_mutex_lock(&cam_mutex);
    if((retval = ssp_initialize(&handle, profile, &settings))!=SSP_SUCCESS){
      
      printf("Can not initialize ssp library. %x\n",retval);
      return;
    }
     
    printf("Initialize done.\n");


    ssp_control_focus_on_camera(handle);
    printf("start streaming\n");
    if((retval = ssp_start_streaming(handle)) != SSP_SUCCESS){
	
      printf("Can not start streaming. %x\n",retval);
      return;
    }
    pthread_mutex_unlock(&cam_mutex);

    struct timeval tv_start, tv_end;
    struct timezone tz_local;
    gettimeofday(&tv_start, &tz_local);
    
    cvWaitKey(2);

    while(1){ if(stop_process==1) break;}
      
    gettimeofday(&tv_end, &tz_local);

    double start_time = tv_start.tv_sec+(double)tv_start.tv_usec/1000000;
    double end_time = tv_end.tv_sec+(double)tv_end.tv_usec/1000000;
    double fps;
    fps = (double)frame_count[cam_param->CamNo]/(end_time-start_time);
    printf("FrameRate = %f\n", fps);
    
    printf("FrameCount = %d\n", frame_count[cam_param->CamNo]);
    printf("stop streaming\n");
    pthread_mutex_lock(&cam_mutex);
    if((retval = ssp_stop_streaming(handle)) != SSP_SUCCESS){
	
      printf("Can not start streaming. %x\n",retval);
      return;
    }
    pthread_mutex_unlock(&cam_mutex);
    
    printf("Flush Wait.\n");
    ssp_flush_event(handle);
    printf("Flush Done.\n");

    if(ssp_finalize(handle) != SSP_SUCCESS){
      printf("Error during finalization.\n");
      return;
    }
    
    cvReleaseImage(&RGBImg[cam_param->CamNo]);
    cvReleaseImage(&BayerImg[cam_param->CamNo]);
    if(cam_param->CamNo == ActiveCamNo)
      cvDestroyWindow(WindowName);
    
    printf("Finalize done.\n");fflush(stdout);
  }
}


int main(int argc, char **argv){

  if(argc < 3){
    printf("SSP library viewer sample program with mutiple camera\n");
    printf("Copyright (c) Streamtechnology Corporation.\n");
    printf("              http://www.streamtechnology.co.jp/\n");
    printf("Usage: %s profile_file_Cam0 model_Cam0 profile_Cam1 model_Cam1\n", argv[0]);
    printf("\tmodel: IMX219 or IMX408 or IU233\n");
    return 0;
  }

  model_name[0] = argv[2];
  model_name[1] = argv[4];

  char *profile_filename[2];

  profile_filename[0] = argv[1];
  profile_filename[1] = argv[3];
  
  stop_process = 0;
  if (signal(SIGINT, signal_handler) == SIG_ERR)
    return 2;

  int AvailableNumCameras;
  if(ssp_get_available_cameras (&AvailableNumCameras) != SSP_SUCCESS){
    printf("Error: Can not get available number of cameras.\n");
    return 0;
  }

  printf("Available number of cameras = %d\n", AvailableNumCameras);
  if(AvailableNumCameras < 2){
    printf("Error: this applcation needs two Cameras.%d\n");
    return 0;
  }
  
  frame_count[0] = 0;
  frame_count[1] = 0;

  pthread_mutex_init(&cam_mutex, NULL);
  pthread_mutex_init(&frame_mutex, NULL);
  
  pthread_t cam_thread[2];
  struct __cam_thread_args cam_params[2];


  cam_params[0].CamNo = 0;
  cam_params[0].profile_filename = profile_filename[0];
  cam_params[0].model_name = model_name[0];
  
  pthread_create(&(cam_thread[0]) , NULL , cam_thread_func , &cam_params[0]);
  usleep(1000000);
  
  cam_params[1].CamNo = 1;
  cam_params[1].profile_filename = profile_filename[1];
  cam_params[1].model_name = model_name[1];

  pthread_create(&(cam_thread[1]) , NULL , cam_thread_func , &cam_params[1]);


  while(stop_process == 0){}

  pthread_join(cam_thread[0] , NULL);
  pthread_join(cam_thread[1] , NULL);

  pthread_mutex_destroy(&cam_mutex);
  return 1;
}
